#include <vector>
#include <limits>

#include <landscape.h>
#include <interfaces/IFontMan.h>
#include <modules/actors/simpleactor.h>
#include <modules/scripting/mappings.h>
#include <sound.h>

#include "gunsight.h"

struct MissileWarningModule : public GunsightModule, public SigObject {
    std::vector<Ptr<IActor> > missiles;
	Ptr<SimpleActor> actor;
	Ptr<IFontMan> fontman;
    Ptr<SoundMan> soundman;
    Ptr<SoundSource> sndsource;
	
	MissileWarningModule(const char *name, Ptr<IGame> game, Ptr<SimpleActor> actor)
	:	GunsightModule(name, 200, 25),
		actor(actor),
        fontman(game->getFontMan()),
        soundman(game->getSoundMan())
	{
        actor->message_signal.connect(
            SigC::slot(*this, &MissileWarningModule::onMessage));
        sndsource = soundman->requestSource();
        sndsource->setLooping(true);
    }

    void onMessage(std::string name, IoObject *args) {
        if (name != "missileShot")
            return;

        ls_message("MissileWarningModule: Handling missileShot event\n");

        IoObject *self = args;
        IoObject *io_missile = IoObject_rawGetSlot_(args, IOSYMBOL("missile"));
        if (!io_missile)
            return;

        Ptr<IActor> missile = unwrapObject<Ptr<IActor> >(io_missile);

        missiles.push_back(missile);

        if (!sndsource->isPlaying()) {
            sndsource->play(soundman->querySound("missile-warning.wav"));
            sndsource->setLooping(true);
        }
    }

    void updateMissileList() {
	    int removed=0;
	    for(int i=0; i<missiles.size(); ++i) {
		    Ptr<IActor> a=missiles[i];
            if (!a->isAlive()) {
			    removed++;
		    } else if (removed>0) missiles[i-removed]=a;
	    }
	    missiles.resize(missiles.size()-removed);
        if (sndsource->isPlaying() && missiles.empty())
            sndsource->setLooping(false);
    }

    std::pair<Ptr<IActor>, float> closestMissile() {
        typedef std::vector<Ptr<IActor> >::iterator Iter;
        Ptr<IActor> closest;
        float min_dist = std::numeric_limits<float>::infinity();
        for(Iter i=missiles.begin(); i!= missiles.end(); ++i) {
            float dist = ((*i)->getLocation() - actor->getLocation()).length();
            if (dist < min_dist) {
                min_dist = dist;
                closest = *i;
            }
        }
        return std::make_pair(closest, min_dist);
    }
	
	void draw(FlexibleGunsight & gunsight) {
        updateMissileList();

        std::pair<Ptr<IActor>, float> closest = closestMissile();
        if (!closest.first) return;

        float pitch = 2.0f*3000 / (closest.second + 3000);

        sndsource->setPosition(actor->getLocation());
        sndsource->setPitch(pitch);

		UI::Surface surface = gunsight.getSurface();
		surface.translateOrigin(offset[0],offset[1]);
		
		fontman->selectFont(IFontMan::FontSpec(
            "dungeon", 12, IFontMan::FontSpec::BOLD));
		
		fontman->setCursor(
			surface.getOrigin(),
			surface.getDX(),
			surface.getDY());
		fontman->setAlpha(1);
		fontman->setColor(Vector(1,0.2,0));
		
		char buf[256];
        snprintf(buf,256,"MISSILE WARNING\n", closest.second);
		fontman->print(buf);
	}
};

void FlexibleGunsight::addMissileWarning(
	Ptr<IGame> game, Ptr<SimpleActor> actor)
{
    addModule(new MissileWarningModule("missile-warning", game, actor),
        "screen", HCENTER | BOTTOM, HCENTER | BOTTOM, Vector(0,5,0));
}
